@@ -60,8 +60,7 @@ public:
 			strandFix_(strandFix),
 			rangeMode_(rangeMode),
 			verbose_(verbose),
-			quiet_(quiet),
-			seed_(seed)
+			quiet_(quiet)
 	{
 		assert(ebwtFw.isInMemory());
 		assert(ebwtBw != NULL);
@@ -240,7 +239,6 @@ private:
 	const bool rangeMode_;
 	const bool verbose_;
 	const bool quiet_;
-	uint32_t seed_;
 };
 
 /**
@@ -312,8 +310,7 @@ public:
 			strandFix_(strandFix),
 			rangeMode_(rangeMode),
 			verbose_(verbose),
-			quiet_(quiet),
-			seed_(seed)
+			quiet_(quiet)
 	{
 		assert(ebwtBw != NULL);
 		assert(ebwtFw.isInMemory());
@@ -685,7 +682,6 @@ private:
 	const bool rangeMode_;
 	const bool verbose_;
 	const bool quiet_;
-	const uint32_t seed_;
 };
 
 #endif /* ALIGNER_23MM_H_ */
