@@ -15,6 +15,21 @@
 #include <stdexcept>
 #include "assert_helpers.h"
 
+/**
+ * Simple, fast helper for determining if a character is a newline.
+ */
+static inline bool isnewline(int c) {
+	return c == '\r' || c == '\n';
+}
+
+/**
+ * Simple, fast helper for determining if a character is a non-newline
+ * whitespace character.
+ */
+static inline bool isspace_notnl(int c) {
+	return isspace(c) && !isnewline(c);
+}
+
 /**
  * Simple wrapper for a FILE*, istream or ifstream that reads it in
  * chunks (with fread) and keeps those chunks in a buffer.  It also
@@ -23,6 +38,7 @@
  */
 class FileBuf {
 public:
+
 	FileBuf() {
 		init();
 	}
@@ -45,6 +61,9 @@ public:
 		assert(_ins != NULL);
 	}
 
+	/**
+	 * Return true iff there is a stream ready to read.
+	 */
 	bool isOpen() {
 		return _in != NULL || _inf != NULL || _ins != NULL;
 	}
@@ -161,6 +180,7 @@ public:
 					_buf_sz = _ins->gcount();
 				} else {
 					assert(_in != NULL);
+					// TODO: consider an _unlocked function
 					_buf_sz = fread(_buf, 1, BUF_SZ, _in);
 				}
 				_cur = 0;
@@ -192,12 +212,12 @@ public:
 				buf[stored] = '\0';
 				return stored;
 			}
-			if(stored == len-1 || c == '\n' || c == '\r') {
+			if(stored == len-1 || isnewline(c)) {
 				// End of string
 				buf[stored] = '\0';
 				// Skip over all end-of-line characters
 				int pc = peek();
-				while(pc == '\n' || pc == '\r') {
+				while(isnewline(pc)) {
 					get(); // discard
 					pc = peek();
 				}
@@ -242,8 +262,8 @@ public:
 	 */
 	int getPastNewline() {
 		int c = get();
-		while(c != '\r' && c != '\n' && c != -1) c = get();
-		while(c == '\r' || c == '\n') c = get();
+		while(!isnewline(c) && c != -1) c = get();
+		while(isnewline(c)) c = get();
 		assert_neq(c, '\r');
 		assert_neq(c, '\n');
 		return c;
@@ -256,8 +276,8 @@ public:
 	 */
 	int peekPastNewline() {
 		int c = peek();
-		while(c != '\r' && c != '\n' && c != -1) c = get();
-		while(c == '\r' || c == '\n') c = get();
+		while(!isnewline(c) && c != -1) c = get();
+		while(isnewline(c)) c = get();
 		assert_neq(c, '\r');
 		assert_neq(c, '\n');
 		return c;
@@ -269,10 +289,10 @@ public:
 	 */
 	int peekUptoNewline() {
 		int c = peek();
-		while(c != '\r' && c != '\n' && c != -1) {
+		while(!isnewline(c) && c != -1) {
 			get(); c = peek();
 		}
-		while(c == '\r' || c == '\n') {
+		while(isnewline(c)) {
 			get();
 			c = peek();
 		}
@@ -309,7 +329,7 @@ public:
 	/**
 	 * Get current size of the last-N-chars buffer.
 	 */
-	const size_t lastNLen() const {
+	size_t lastNLen() const {
 		return _lastn_cur;
 	}
 
@@ -412,6 +432,23 @@ class OutFileBuf {
 
 public:
 
+    static const size_t BUF_SZ = 16 * 1024;
+
+	/**
+	 * Open a new output stream to a file with given name.
+	 */
+	OutFileBuf(const std::string& out, bool binary = false) :
+		name_(out.c_str()), cur_(0), closed_(false)
+	{
+		out_ = fopen(out.c_str(), binary ? "wb" : "w");
+		if(out_ == NULL) {
+			std::cerr << "Error: Could not open alignment output file " << out.c_str() << std::endl;
+			throw 1;
+		}
+		if(setvbuf(out_, NULL, _IOFBF, 10* 1024* 1024))
+			std::cerr << "Warning: Could not allocate the proper buffer size for output file stream. " << std::endl;
+	}
+
 	/**
 	 * Open a new output stream to a file with given name.
 	 */
@@ -435,6 +472,11 @@ public:
 		out_ = stdout;
 	}
 
+	/**
+	 * Close buffer when object is destroyed.
+	 */
+	~OutFileBuf() { close(); }
+
 	/**
 	 * Open a new output stream to a file with given name.
 	 */
@@ -486,6 +528,35 @@ public:
 		assert_leq(cur_, BUF_SZ);
 	}
 
+	/**
+	 * Write a c++ string to the write buffer and, if necessary, flush.
+	 */
+	template<typename T>
+	size_t writeString(const T& s) {
+		assert(!closed_);
+		size_t slen = s.length();
+		size_t bytes_written = 0;
+		if(cur_ + slen > BUF_SZ) {
+			if(cur_ > 0) {
+				flush();
+				bytes_written += cur_;;
+			}
+			if(slen >= BUF_SZ) {
+				fwrite(s.toZBuf(), slen, 1, out_);
+			} else {
+				memcpy(&buf_[cur_], s.toZBuf(), slen);
+				assert_eq(0, cur_);
+				cur_ = slen;
+			}
+		} else {
+			memcpy(&buf_[cur_], s.toZBuf(), slen);
+			cur_ += slen;
+		}
+		bytes_written += slen;
+		assert_leq(cur_, BUF_SZ);
+		return bytes_written;
+	}
+
 	/**
 	 * Write a c++ string to the write buffer and, if necessary, flush.
 	 */
@@ -513,6 +584,13 @@ public:
 		assert_leq(cur_, BUF_SZ);
 	}
 
+	/**
+	 * Write a 0-terminated C string to the output stream.
+	 */
+	void writeChars(const char * s) {
+		writeChars(s, strlen(s));
+	}
+
 	/**
 	 * Write any remaining bitpairs and then close the input
 	 */
@@ -557,8 +635,6 @@ public:
 
 private:
 
-	static const size_t BUF_SZ = 16 * 1024;
-
 	const char *name_;
 	FILE       *out_;
 	size_t    cur_;
