@@ -4,6 +4,7 @@
 #include <stdexcept>
 #include <seqan/sequence.h>
 #include <seqan/file.h>
+#include <string.h>
 
 #include "pat.h"
 #include "filebuf.h"
@@ -226,21 +227,16 @@ pair<bool, int> DualPatternComposer::nextBatch(PerThreadReadBuf& pt) {
  * Returns pair<bool, int> where bool indicates whether we're
  * completely done, and int indicates how many reads were read.
  */
-pair<bool, int> CFilePatternSource::nextBatch(
+pair<bool, int> CFilePatternSource::nextBatchImpl(
 	PerThreadReadBuf& pt,
-	bool batch_a,
-	bool lock)
+	bool batch_a)
 {
 	bool done = false;
-	int nread = 0;
-	
-	// synchronization at this level because both reading and manipulation of
-	// current file pointer have to be protected
-	ThreadSafe ts(&mutex, lock);
+	size_t nread = 0;
 	pt.setReadId(readCnt_);
 	while(true) { // loop that moves on to next file when needed
 		do {
-			pair<bool, int> ret = nextBatchFromFile(pt, batch_a);
+			pair<bool, int> ret = nextBatchFromFile(pt, batch_a, nread);
 			done = ret.first;
 			nread = ret.second;
 		} while(!done && nread == 0); // not sure why this would happen
@@ -248,9 +244,10 @@ pair<bool, int> CFilePatternSource::nextBatch(
 			open();
 			resetForNextFile(); // reset state to handle a fresh file
 			filecur_++;
-			if(nread == 0) {
+			if(nread == 0 || nread < pt.max_buf_) {
 				continue;
 			}
+			done = false;
 		}
 		break;
 	}
@@ -259,15 +256,36 @@ pair<bool, int> CFilePatternSource::nextBatch(
 	return make_pair(done, nread);
 }
 
+pair<bool, int> CFilePatternSource::nextBatch(
+	PerThreadReadBuf& pt,
+	bool batch_a,
+	bool lock)
+{
+	if(lock) {
+		// synchronization at this level because both reading and manipulation of
+		// current file pointer have to be protected
+		ThreadSafe ts(&mutex);
+		return nextBatchImpl(pt, batch_a);
+	} else {
+		return nextBatchImpl(pt, batch_a);
+	}
+}
+
 /**
  * Open the next file in the list of input files.
  */
 void CFilePatternSource::open() {
 	if(is_open_) {
 		is_open_ = false;
-		fclose(fp_);
-		fp_ = NULL;
-		if(qfp_ != NULL) {
+		if (compressed_) {
+			gzclose(zfp_);
+			zfp_ = NULL;
+		}
+		else if (fp_ != stdin) {
+			fclose(fp_);
+			fp_ = NULL;
+		}
+		if(qfp_ != NULL && qfp_ != stdin) {
 			fclose(qfp_);
 			qfp_ = NULL;
 		}
@@ -275,19 +293,42 @@ void CFilePatternSource::open() {
 	while(filecur_ < infiles_.size()) {
 		// Open read
 		if(infiles_[filecur_] == "-") {
-			fp_ = stdin;
-		} else if((fp_ = fopen(infiles_[filecur_].c_str(), "rb")) == NULL) {
-			if(!errs_[filecur_]) {
-				cerr << "Warning: Could not open read file \""
-				     << infiles_[filecur_] << "\" for reading; skipping..."
-					 << endl;
-				errs_[filecur_] = true;
+			compressed_ = true;
+			int fn = dup(fileno(stdin));
+			zfp_ = gzdopen(fn, "rb");
+		}
+		else {
+			compressed_ = false;
+			if (is_gzipped_file(infiles_[filecur_])) {
+				compressed_ = true;
+				zfp_ = gzopen(infiles_[filecur_].c_str(), "rb");
+			}
+			else {
+				fp_ = fopen(infiles_[filecur_].c_str(), "rb");
+			}
+			if ((compressed_ && zfp_ == NULL) || (!compressed_ && fp_ == NULL)) {
+				if(!errs_[filecur_]) {
+					cerr << "Warning: Could not open read file \""
+					     << infiles_[filecur_] << "\" for reading; skipping..."
+					     << endl;
+					errs_[filecur_] = true;
+				}
+				filecur_++;
+				continue;
 			}
-			filecur_++;
-			continue;
 		}
 		is_open_ = true;
-		setvbuf(fp_, buf_, _IOFBF, 64*1024);
+		if (compressed_) {
+#if ZLIB_VERNUM < 0x1235
+			cerr << "Warning: gzbuffer added in zlib v1.2.3.5. Unable to change "
+			        "buffer size from default of 8192." << endl;
+#else
+			gzbuffer(zfp_, 64*1024);
+#endif
+		}
+		else {
+			setvbuf(fp_, buf_, _IOFBF, 64*1024);
+		}
 		if(!qinfiles_.empty()) {
 			if(qinfiles_[filecur_] == "-") {
 				qfp_ = stdin;
@@ -317,10 +358,9 @@ void CFilePatternSource::open() {
 VectorPatternSource::VectorPatternSource(
 	const vector<string>& seqs,
 	bool color,
-	const char *dumpfile,
 	int trim3,
 	int trim5) :
-	TrimmingPatternSource(dumpfile, trim3, trim5),
+	TrimmingPatternSource(trim3, trim5),
 	color_(color),
 	cur_(0),
 	paired_(false),
@@ -364,12 +404,10 @@ VectorPatternSource::VectorPatternSource(
  * in the contsructor.  This essentially modifies the pt as though we read
  * in some number of patterns.
  */
-pair<bool, int> VectorPatternSource::nextBatch(
+pair<bool, int> VectorPatternSource::nextBatchImpl(
 	PerThreadReadBuf& pt,
-	bool batch_a,
-	bool lock)
+	bool batch_a)
 {
-	ThreadSafe ts(&mutex, lock);
 	pt.setReadId(cur_);
 	vector<Read>& readbuf = batch_a ? pt.bufa_ : pt.bufb_;
 	size_t readi = 0;
@@ -382,6 +420,19 @@ pair<bool, int> VectorPatternSource::nextBatch(
 	return make_pair(cur_ == bufs_.size(), readi);
 }
 
+pair<bool, int> VectorPatternSource::nextBatch(
+	PerThreadReadBuf& pt,
+	bool batch_a,
+	bool lock)
+{
+	if(lock) {
+		ThreadSafe ts(&mutex);
+		return nextBatchImpl(pt, batch_a);
+	} else {
+		return nextBatchImpl(pt, batch_a);
+	}
+}
+
 /**
  * Finishes parsing outside the critical section.
  */
@@ -522,14 +573,18 @@ bool VectorPatternSource::parse(Read& ra, Read& rb, TReadId rdid) const {
  */
 pair<bool, int> FastaPatternSource::nextBatchFromFile(
 	PerThreadReadBuf& pt,
-	bool batch_a)
+	bool batch_a,
+	size_t readi)
 {
 	int c;
 	vector<Read>& readbuf = batch_a ? pt.bufa_ : pt.bufb_;
 	if(first_) {
-		c = getc_unlocked(fp_);
+		c = getc_wrapper();
+		if(c == EOF) {
+			return make_pair(true, 0);
+		}
 		while(c == '\r' || c == '\n') {
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 		}
 		if(c != '>') {
 			cerr << "Error: reads file does not look like a FASTA file" << endl;
@@ -538,13 +593,12 @@ pair<bool, int> FastaPatternSource::nextBatchFromFile(
 		first_ = false;
 	}
 	bool done = false;
-	size_t readi = 0;
 	// Read until we run out of input or until we've filled the buffer
 	for(; readi < pt.max_buf_ && !done; readi++) {
 		readbuf[readi].readOrigBuf[0] = '>';
 		size_t bufoff = 1;
 		while(true) {
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 			if(c < 0 || c == '>') {
 				done = c < 0;
 				break;
@@ -553,6 +607,10 @@ pair<bool, int> FastaPatternSource::nextBatchFromFile(
 		}
 		readbuf[readi].readOrigBufLen = bufoff;
 	}
+	// Immediate EOF case
+	if(done && readbuf[readi-1].readOrigBufLen == 1) {
+		readi--;
+	}
 	return make_pair(done, readi);
 }
 
@@ -682,21 +740,21 @@ bool FastaPatternSource::parse(Read& r, Read& rb, TReadId rdid) const {
  */
 pair<bool, int> FastaContinuousPatternSource::nextBatchFromFile(
 	PerThreadReadBuf& pt,
-	bool batch_a)
+	bool batch_a,
+	size_t readi)
 {
 	int c = -1;
 	vector<Read>& readbuf = batch_a ? pt.bufa_ : pt.bufb_;
-	size_t readi = 0;
 	int nameoff = 0;
 	while(readi < pt.max_buf_) {
-		c = getc_unlocked(fp_);
+		c = getc_wrapper();
 		if(c < 0) {
 			break;
 		}
 		if(c == '>') {
 			resetForNextFile();
 			nameoff = 0;
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 			bool sawSpace = false;
 			while(c != '\n' && c != '\r') {
 				if(!sawSpace) {
@@ -708,10 +766,10 @@ pair<bool, int> FastaContinuousPatternSource::nextBatchFromFile(
 					// that are substrings of this FASTA sequence
 					name_prefix_buf_[nameoff++] = c;
 				}
-				c = getc_unlocked(fp_);
+				c = getc_wrapper();
 			}
 			while(c == '\n' || c == '\r') {
-				c = getc_unlocked(fp_);
+				c = getc_wrapper();
 			}
 			if(c < 0) {
 				break;
@@ -846,32 +904,32 @@ bool FastaContinuousPatternSource::parse(
  */
 pair<bool, int> FastqPatternSource::nextBatchFromFile(
 	PerThreadReadBuf& pt,
-	bool batch_a)
+	bool batch_a,
+	size_t readi)
 {
 	int c = 0;
-	vector<Read>& readBuf = batch_a ? pt.bufa_ : pt.bufb_;
+	vector<Read>* readBuf = batch_a ? &pt.bufa_ : &pt.bufb_;
 	if(first_) {
-		c = getc_unlocked(fp_);
+		c = getc_wrapper();
 		while(c == '\r' || c == '\n') {
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 		}
 		if(c != '@') {
 			cerr << "Error: reads file does not look like a FASTQ file" << endl;
 			throw 1;
 		}
 		first_ = false;
-		readBuf[0].readOrigBuf[0] = c;
-		readBuf[0].readOrigBufLen = 1;
+		(*readBuf)[readi].readOrigBuf[0] = c;
+		(*readBuf)[readi].readOrigBufLen = 1;
 	}
 	bool done = false, aborted = false;
-	size_t readi = 0;
 	// Read until we run out of input or until we've filled the buffer
-	for(; readi < pt.max_buf_ && !done; readi++) {
-		char* buf = readBuf[readi].readOrigBuf;
-		assert(readi == 0 || readBuf[readi].readOrigBufLen == 0);
+	while (readi < pt.max_buf_ && !done) {
+		char* buf = (*readBuf)[readi].readOrigBuf;
+		assert(readi == 0 || (*readBuf)[readi].readOrigBufLen == 0);
 		int newlines = 4;
 		while(newlines) {
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 			done = c < 0;
 			if(c == '\n' || (done && newlines == 1)) {
 				// Saw newline, or EOF that we're
@@ -879,10 +937,26 @@ pair<bool, int> FastqPatternSource::nextBatchFromFile(
 				newlines--;
 				c = '\n';
 			} else if(done) {
-				aborted = true; // Unexpected EOF
+				if (newlines == 4) {
+					newlines = 0;
+				} else {
+					aborted = true; // Unexpected EOF
+				}
 				break;
 			}
-			buf[readBuf[readi].readOrigBufLen++] = c;
+			buf[(*readBuf)[readi].readOrigBufLen++] = c;
+		}
+		if (c > 0) {
+			if (interleaved_) {
+				// alternate between read buffers
+				batch_a = !batch_a;
+				readBuf = batch_a ? &pt.bufa_ : &pt.bufb_;
+				// increment read counter after each pair gets read
+				readi = batch_a ? readi + 1 : readi;
+			}
+			else {
+				readi++;
+			}
 		}
 	}
 	if(aborted) {
@@ -1029,8 +1103,18 @@ bool FastqPatternSource::parse(Read &r, Read& rb, TReadId rdid) const {
 			tooFewQualities(r.name);
 			return false;
 		} else if(qualoff > seqoff) {
-			tooManyQualities(r.name);
-			return false;
+			// if qualoff is at most 2 characters longer than the sequence
+			// then the extra characters will most likely be the quality values
+			// of the primer and the first base (which get discarded by bowtie).
+			// In this case move the remainder of the sequence the (qualoff - seqoff)
+			// positions left.
+			if (r.color && qualoff - seqoff <= 2) {
+				memmove(r.qualBuf, r.qualBuf + (qualoff - seqoff), seqoff);
+			}
+			else {
+				tooManyQualities(r.name);
+				return false;
+			}
 		}
 	}
 	r.qualBuf[seqan::length(r.patFw)] = '\0';
@@ -1055,23 +1139,23 @@ bool FastqPatternSource::parse(Read &r, Read& rb, TReadId rdid) const {
  */
 pair<bool, int> TabbedPatternSource::nextBatchFromFile(
 	PerThreadReadBuf& pt,
-	bool batch_a)
+	bool batch_a,
+	size_t readi)
 {
-	int c = getc_unlocked(fp_);
+	int c = getc_wrapper();
 	while(c >= 0 && (c == '\n' || c == '\r')) {
-		c = getc_unlocked(fp_);
+		c = getc_wrapper();
 	}
 	vector<Read>& readbuf = batch_a ? pt.bufa_ : pt.bufb_;
-	size_t readi = 0;
 	// Read until we run out of input or until we've filled the buffer
 	for(; readi < pt.max_buf_ && c >= 0; readi++) {
 		readbuf[readi].readOrigBufLen = 0;
 		while(c >= 0 && c != '\n' && c != '\r') {
 			readbuf[readi].readOrigBuf[readbuf[readi].readOrigBufLen++] = c;
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 		}
-		while(c >= 0 && (c == '\n' || c == '\r')) {
-			c = getc_unlocked(fp_);
+		while(c >= 0 && (c == '\n' || c == '\r') && readi < pt.max_buf_ - 1) {
+			c = getc_wrapper();
 		}
 	}
 	return make_pair(c < 0, readi);
@@ -1234,23 +1318,23 @@ bool TabbedPatternSource::parse(Read& ra, Read& rb, TReadId rdid) const {
  */
 pair<bool, int> RawPatternSource::nextBatchFromFile(
 	PerThreadReadBuf& pt,
-	bool batch_a)
+	bool batch_a,
+	size_t readi)
 {
-	int c = getc_unlocked(fp_);
+	int c = getc_wrapper();
 	while(c >= 0 && (c == '\n' || c == '\r')) {
-		c = getc_unlocked(fp_);
+		c = getc_wrapper();
 	}
 	vector<Read>& readbuf = batch_a ? pt.bufa_ : pt.bufb_;
-	size_t readi = 0;
 	// Read until we run out of input or until we've filled the buffer
 	for(; readi < pt.max_buf_ && c >= 0; readi++) {
 		readbuf[readi].readOrigBufLen = 0;
 		while(c >= 0 && c != '\n' && c != '\r') {
 			readbuf[readi].readOrigBuf[readbuf[readi].readOrigBufLen++] = c;
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 		}
 		while(c >= 0 && (c == '\n' || c == '\r')) {
-			c = getc_unlocked(fp_);
+			c = getc_wrapper();
 		}
 	}
 	return make_pair(c < 0, readi);
